!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Original module written by Julien Lesgourgues for using weak lensing data in CosmoMC        !
! It was written for the COSMOS 3D data by Massey et al. [astro-ph/0701480],                  !
! was used in [arXiv:0705.0533] by J. Lesgourgues, M. Viel, M.G. Haehnelt, R. Massey          !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This modified version was written by Jason Dossett.                                         !
! It allows the use to the refined Cosmos3D data by Schrabback et al [arXiv:0911.0053],       !
! and was first used in [arXiv:1103.1195] by J. Dossett, J. Moldenhauer, M. Ishak             !
! to probe deviations from General Relativity.                                                !
! This version makes no assumptions about the curvature and is consistent with                !
! formulae used in [arXiv:1205.2422]                                                          !
! Last Modified 05/2012                                                                       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

MODULE WeakLen
  use Camb, only : f_k_real
  use cmbtypes
  use AngularS ! ancillary module converting harmonic power spectrum into angular correlation functions
  implicit none
  real(dl), parameter :: piwl = 3.1415926535897932384626433832795_dl
  ! precision parameters:
  integer, parameter :: nlmax=130  ! fixes the maximum required l value in C_l(lensing) 
  real, parameter :: dlnl=0.1      ! fixes logarithmic l step for C_l(lensing)
  real, parameter :: xstop=200.    ! fixes maximum value of x =  l * theta in angular correlation function 
  ! note: integration of C_l J_0,4(l*theta) over l will stop 
  ! at l = min( ll(nlmax),  xstop/theta )

  ! parameters describing Schrabback's refined Cosmos3D data set:
  
  integer, parameter :: nbinmax=6  ! Number of redshift bins
    
  integer, parameter :: nzmax = 50 ! Number of z values in galaxy distribution file
   
  integer, parameter :: nz6 = 500  ! Number of z values in data file for 
                                     ! the estimiated galaxy distribution of the sixth bin
  
  integer, parameter :: nthetamax=5 ! number of angles theta
  integer, parameter :: ncl = 16    ! number of Cross-power spectra to compute
   
  logical :: Initialize_WeakLen = .false.
!Data Arrays
  real z(nzmax),etaz(nzmax,nbinmax),eta_norm(nbinmax) ! data's selection functions
  real z6(nz6),etaz6(nz6)  ! Allows for computation of 6th bin's galaxy distribution with 
                           ! Nuissance parameter described in [arXiv:0911.0053]
  real(dl) theta(nthetamax) ! data array of theta values
  !Arrays for calculating -ln(Like)
  real d_theo(2*nthetamax*ncl)   ! Theoretical shear cross-correlations, data arranged like a vector
  real d_obs(2*nthetamax*ncl)  ! Observed shear cross-correlations, data arranged like a vector
  real invcov(2*nthetamax*ncl,2*nthetamax*ncl)  !inverse Covariance Matrix

!Below for weighting of cross-correlations as described in [arXiv:0911.0053] and []  
  integer, parameter :: nthetastep = 100 ! for number of integration steps 
  real(dl) dtheta(nthetamax) ! dtheta for each theta bin given number of integration steps
  real(dl) pownorm(nthetamax) !for each theata bin Int_\thetabinmin^thetabinmax  N(\theta) d\theta
  
CONTAINS

  FUNCTION WeakLenLnLike(CMB, Theory)
    TYPE (CMBParams) CMB
    TYPE (CosmoTheory) Theory
    REAL :: WeakLenLnLike
    
    integer nl,nc,clc1,clc2
    real ll(nlmax)
    integer nkmax,nk
    integer zix
    real kk(num_power_spec-1)
    real P(num_power_spec-1,power_spec_lnzsteps)    ! P(k,z)
    real ddP(num_power_spec-1,power_spec_lnzsteps)  ! same splined along k
    real PP(power_spec_lnzsteps)                      ! P(k=l/r, z=z(r))
    real ddPP(power_spec_lnzsteps)                    ! same splined along r
    ! for spline:
    real sig_spline,p_spline,u_spline(num_power_spec-1)
    ! for splint:
    integer k,khi,klo
    real a,b,h
    ! for eta(z):
    integer nz,nbin,nr,nrmax,nrs
    real r(nzmax),dz_o_dr(nzmax),etar(nzmax,nbinmax)
    real g(nzmax,nbinmax)
    character*100 name
    ! for Theory%z_and_r(:,zix)
    real zz(power_spec_lnzsteps),rr(power_spec_lnzsteps),dz_over_dr(power_spec_lnzsteps)
    real ddrr(power_spec_lnzsteps),dddz_over_dr(power_spec_lnzsteps)
    ! For C_l
    real kvalue,PP_r
    real Cl_integrand(nzmax,ncl)
    real C(nlmax,ncl),ddC(nlmax,ncl)
    real b1, b2  !Dummies for reading in correlation vector
    !JD for getting new values for bin 6 with nuissance parameter involved.
    real z62(nz6),ddetaz6(nz6)
    !JD for integration of pownorm
    real(dl) thetamax
    real(dl) normint(4,nthetamax) ! For Simpsons rule integreation
    real(dl) theta2(nthetamax),radtomin
    integer t
    !For C(theta)
    real Cp(nthetamax,ncl),Cm(nthetamax,ncl) !cross-correlation arrays for function Cl to Ctheta
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Initialization !!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    if (Initialize_Weaklen .eq. .false.) then

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Read data !!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if(feedback>1)write(*,*)'Initializing refined COSMOS 3D weak lensing data'


        !Read galaxy distribution for bins 1 to 5
        name='data/COSMOS_Schrabback2010_zdist_bin1to5JD.txt'
        open (unit=10,file=name)
        do nz=1,nzmax
            read(10,*)z(nz), etaz(nz,:)
        end do
        close(10)

        !Read galaxy distribution for bin 6
        name='data/COSMOS_Schrabback2010_zdist_bin6_fit.txt'
        open (unit=10,file=name)
        do nz=1,nz6
            read(10,*)z6(nz),etaz6(nz)
        end do
        close(10)

        ! Read Thetas
        name='data/COSMOS_Schrabback2010_thetasJD.txt'
        open (unit=10,file=name)
        do k=1,nthetamax
            read(10,*)theta(k),thetamax
            theta(k)=theta(k)/60._dl*piwl/180._dl
            thetamax=thetamax/60._dl*piwl/180._dl
            dtheta(k)=(thetamax-theta(k))/nthetastep   !Set dtheta for each theta bin
        end do
        close(10)
        
        !read data vector
        name='data/COSMOS_Schrabback2010_correlation_vectorJD.txt'
        open (unit=10,file=name)
        read(10,*)name
        do k=1,2*nthetamax*ncl
            read(10,*)b1,b2,d_obs(k)
        end do
        close(10)


        ! Read inverse covariance matrix
        name='data/COSMOS_Schrabback2010_invcovJD.txt'
        open (unit=10,file=name)
        do k=1,2*nthetamax*ncl
            read(10,*)invcov(k,:)
        end do
        close(10)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! integrate eta(z) over z (in view of normalizing it to one) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        eta_norm(:)=0.
        do nz=2,nzmax
            eta_norm(:)=eta_norm(:)+0.5*(etaz(nz-1,:)+etaz(nz,:))*(z(nz)-z(nz-1))
        end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Compute Bessel interpolation array !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        call GetBessels(xstop*1.d0) ! subroutine of the "Angular" module

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!! Perform normalization integral                 !!!!!
!!!! \int_{thetaxmin}^{thetaxmax} N(\theta) d\theta !!!!!
!!!! Via Simpson's 3/8's rule                       !!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        radtomin = 60._dl*180._dl/piwl
        t=1
        pownorm(:) = 0._dl
        do k=0,nthetastep-1
            theta2(:)=theta(:)+k*dtheta(:)  !vector of thetas for each step
            !fill in vector of steps
            normint(t,:)= radtomin*theta2(:)*(0.0004664+radtomin*theta2(:)&
                    *(0.0044118-8.90878e-5*radtomin*theta2(:)))
            if(t==4)then
                pownorm(:)= pownorm(:)&
                    +(3._dl/8._dl*(normint(1,:)+normint(4,:))&
                    +9._dl/8._dl*(normint(2,:)+normint(3,:)))*dtheta(:)
                normint(1,:)=normint(4,:)
                t=2
            else
                t=t+1
            end if
        end do

        Initialize_Weaklen = .true.

    end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! End of initialization !!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!! Various steps for Likelihood computation start here !!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! JD inclusion of Schrabback nuissance parameter !!!!!!!!
!!!!!!!! this parameter rescales the galaxy redshift !!!!!!!!!!!  
!!!!!!!! distrobution of bin 6 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    z62 = z6/CMB%norm(norm_fz)
    !JD prepare for cubic splin interpolation interpolation
    call spline_real(z62,etaz6,nz6,ddetaz6)
    eta_norm(6) = 0.
    
    ! for each value of z in the data's selection function file, use adjusted data from bin 6,
    ! and get a data point in bin 6 at the correct z
    ! (inspired from Numerical Recipies's splint)
    nrmax=nzmax
    do nr=1,nzmax
        klo=1
        khi=nz6
        do while (khi-klo.gt.1)
            k=(khi+klo)/2
            if(z62(k).gt.z(nr)) then
                khi=k
            else
                klo=k
            endif
        end do
        h=z62(khi)-z62(klo)
        if (h.eq.0.) then
            write(*,*)'Problem: h=0 in splint of etaz6, for nr=',nr,' z(nr)=',z(nr)
            write(*,*)'while klo=',klo,' z62(klo)=',z62(klo),' khi=',khi,' z62(khi)=',z62(khi)
            stop
        end if
        a=(z62(khi)-z(nr))/h
        b=(z(nr)-z62(klo))/h
        etaz(nr,6)=a*etaz6(klo)+b*etaz6(khi)+((a**3-a)*ddetaz6(klo)&
            +(b**3-b)*ddetaz6(khi))*(h**2)/6.
        if(nr>1) eta_norm(6)=eta_norm(6)+0.5*(etaz(nr-1,6)+etaz(nr,6))*(z(nr)-z(nr-1))
    end do


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Convert eta(z) to eta(r) !!!!!!!!!!!!!!!!!!!!
!!! (first unnormalized, second normalized to one) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! array Theory%z_and_r(1,zix) contains list of redshifts zz(zix) in CAMB output 
    ! the first one must be zero:
    if (Theory%z_and_r(1,power_spec_lnzsteps) .eq. 0.) then
        write(*,*)'problem: Theory%z_and_r not passed properly. Results zero in Weak lensing module'
        stop
    end if

    ! read array Theory%z_and_r(1,zix) = list of redshifts zz(zix) in CAMB output 
    ! read array Theory%z_and_r(2,zix) = list of corresponding comoving distances rr(zix)
    ! read array Theory%z_and_r(3,zix) = list of corresponding dz/dr (= Hubble parameter)
    zz(:)=Theory%z_and_r(1,:)  ! note that z (and r, ...) must be rearranged in growing order for splint
    rr(:)=Theory%z_and_r(2,:)
    dz_over_dr(:)=Theory%z_and_r(3,:)


    ! spline rr and dz/dr as functions of zz: then we can get r and dz/dr at any z
    call spline_real(zz,rr,power_spec_lnzsteps,ddrr)
    call spline_real(zz,dz_over_dr,power_spec_lnzsteps,dddz_over_dr)

    ! for each value of z in the data's selection function file, compute r and dz/dr with splint
    ! compute also the selection function eta(r) = eta(z) dz/dr normalized to one
    ! (inspired from Numerical Recipies's splint)
    nrmax=nzmax
    do nr=1,nrmax
        klo=1
        khi=power_spec_lnzsteps
        do while (khi-klo.gt.1)
            k=(khi+klo)/2
            if(zz(k).gt.z(nr)) then
                khi=k
            else
                klo=k
            endif
        end do
        h=zz(khi)-zz(klo)
        if (h.eq.0.) then
            write(*,*)'Problem: h=0 in splint of rr(z), for nr=',nr,' z(nr)=',z(nr)
            write(*,*)'while klo=',klo,' zz(klo)=',zz(klo),' khi=',khi,' zz(khi)=',zz(khi)
            stop
        end if
        a=(zz(khi)-z(nr))/h
        b=(z(nr)-zz(klo))/h
        r(nr)=a*rr(klo)+b*rr(khi)+((a**3-a)*ddrr(klo)+(b**3-b)*ddrr(khi))*(h**2)/6.
        dz_o_dr(nr)=a*dz_over_dr(klo)+b*dz_over_dr(khi)+((a**3-a)*dddz_over_dr(klo)&
            +(b**3-b)*dddz_over_dr(khi))*(h**2)/6.
        etar(nr,:)=etaz(nr,:)*dz_o_dr(nr)/eta_norm(:)      
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! For each bin, compute window function g(r) = r^2 W(r) a_0/a !!!!!!!!!!!
!!!!!!!! with W(r) = 2 sum_0^r_s_max dr_s eta(r_s) (r_a-r)/r/r_s !!!!!!!!!!!!!!!
!!!!!!!! (simple trapezoidal integration ois enough) !!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    g(1,:)=0.
    do nr=2,nrmax-1
        g(nr,:)=0.
        do nrs=nr+1,nrmax
            g(nr,:)=g(nr,:)+0.5*(etar(nrs,:)*f_k_real(r(nrs)-r(nr))/f_k_real(r(nrs)) &
                +etar(nrs-1,:)*f_k_real(r(nrs-1)-r(nr))/f_k_real(r(nrs-1)))*(r(nrs)-r(nrs-1))
        end do
    end do
    g(nrmax,:)=0.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Get power spectrum P(k,z) from CAMB !!!!!!!
!!! and prepare its interpolation w.r.t k !!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! fill array of k
    ! Note that we do a shift of 1 in the argument: 
    ! kk(nk) corresponds to k value (unit h/Mpc) for Theory%lens_power_spec(nk+1,zix)

    kk(:)=Theory%power_spec_k(2:num_power_spec)


    ! fill array of P(k,z)
    ! Note that zz(zix) is arrranged with growing z
    ! this was **already** the case for P(k,zix)
    nkmax=num_power_spec-1
    do zix = 1,power_spec_lnzsteps
        do nk=1,nkmax
            P(nk,zix)= Theory%lens_power_spec(nk+1,zix)
        end do
    end do

    ! for each redshift, spline the P(k,z) w.r.t k
    ! (inspired from Numerical Recipies's spline)
    do zix = 1,power_spec_lnzsteps
        ddP(1,zix)=0.
        u_spline(1)=0.
        do nk=2,nkmax-1
            sig_spline=(kk(nk)-kk(nk-1))/(kk(nk+1)-kk(nk-1))
            p_spline=sig_spline*ddP(nk-1,zix)+2.
            ddP(nk,zix)=(sig_spline-1.)/p_spline
            u_spline(nk)=(6.*((P(nk+1,zix)-P(nk,zix))&
                /(kk(nk+1)-kk(nk))-(P(nk,zix)-P(nk-1,zix))&
                /(kk(nk)-kk(nk-1)))/(kk(nk+1)-kk(nk-1))&
                -sig_spline*u_spline(nk-1))/p_spline
        end do
        ddP(nkmax,zix)=0.
        do nk=nkmax-1,1,-1
            ddP(nk,zix)=ddP(nk,zix)*ddP(nk+1,zix)+u_spline(nk)
        end do
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! start loop over l  for computation of C_l^shear !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nl=1,nlmax

        ! current value of l:
        ll(nl)=1.*exp(dlnl*(nl-1.))

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!! fill array PP(r) = P(l/r,z(r)) !!!!!!!!!!!!!!
        !!! prepare it for interpolation along r  !!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        PP(1)=0.
        do zix = 2,power_spec_lnzsteps
            kvalue=ll(nl)/f_k_real(rr(zix))/CMB%h   ! k=l/r in units h/Mpc
            if ((kvalue .le. kk(1)) .or. (kvalue .ge. kk(nkmax))) then
                PP(zix)=0.
            else
            ! compute P(k,z) at k=l/r
            ! (inspired from Numerical Recipies's splint)
                klo=1
                khi=nkmax
                do while (khi-klo.gt.1)
                    k=(khi+klo)/2
                    if(kk(k).gt.kvalue) then
                       khi=k
                    else
                       klo=k
                    endif
                end do
                h=kk(khi)-kk(klo)
                if (h.eq.0.) then
                    write(*,*)'Problem: h=0 in splint of P(k,z)'
                    stop
                end if
                a=(kk(khi)-kvalue)/h
                b=(kvalue-kk(klo))/h
                PP(zix)=a*P(klo,zix)+b*P(khi,zix)+((a**3-a)*ddP(klo,zix)&
                    +(b**3-b)*ddP(khi,zix))*(h**2)/6.
            end if
        end do
        call spline_real(rr,PP,power_spec_lnzsteps,ddPP)

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! Find PP(r) = P(l/r,z(r)) for each r(nr) !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! and infer Cl_integrand = P^{k l}_\kappa_integrand = (g_{k}(r)*g_{l}(r))*P(l/r,z(r))/r^2 !!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        Cl_integrand(1,:)=0.
        do nr=2,nrmax
        ! (inspired from Numerical Recipies's splint)
            klo=1
            khi=power_spec_lnzsteps
            do while (khi-klo.gt.1)
                k=(khi+klo)/2
                if(rr(k).gt.r(nr)) then
                    khi=k
                else
                    klo=k
                endif
            end do
            h=rr(khi)-rr(klo)
            if (h.eq.0.) then
                write(*,*)'Problem: h=0 in splint of rr(z)'
                stop
            end if
            a=(rr(khi)-r(nr))/h
            b=(r(nr)-rr(klo))/h
            PP_r=a*PP(klo)+b*PP(khi)+((a**3-a)*ddPP(klo)+(b**3-b)*ddPP(khi))*(h**2)/6.
            PP_r=PP_r/(CMB%h)**3  ! in units Mpc**3
            nc =1
            do clc1=1,5 
                do clc2 = clc1+1,6
                    Cl_integrand(nr,nc)=g(nr,clc1)*g(nr,clc2)*PP_r
                    nc = nc+1
                end do
            end do
            if(nc .ne. ncl) then
                write(*,*)'nc count is ',nc, 'after loop instead of ',ncl
                stop
            end if
            Cl_integrand(nr,nc)=g(nr,6)**2*PP_r
        end do

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!!!!!!! Integrate over r to get C_l^shear !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        !!! C_l^shear = 9/16 Omega_m0^2 H_0^4 \sum_0^rmax dr Cl_integrand   !!!!!!!!!
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        C(nl,:)=0.
        do nr=2,nrmax
            C(nl,:)=C(nl,:)+0.5*(Cl_integrand(nr,:)+Cl_integrand(nr-1,:))*(r(nr)-r(nr-1)) 
        end do
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Spline Cl(nl,nbin) along l !!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nbin = 1,ncl
    ! (inspired from Numerical Recipies's spline)
        ddC(1,nbin)=0.
        u_spline(1)=0.
        do nl=2,nlmax-1
            sig_spline=(ll(nl)-ll(nl-1))/(ll(nl+1)-ll(nl-1))
            p_spline=sig_spline*ddC(nl-1,nbin)+2.
            ddC(nl,nbin)=(sig_spline-1.)/p_spline
            u_spline(nl)=(6.*((C(nl+1,nbin)-C(nl,nbin))&
                /(ll(nl+1)-ll(nl))-(C(nl,nbin)-C(nl-1,nbin))&
                /(ll(nl)-ll(nl-1)))/(ll(nl+1)-ll(nl-1))&
                -sig_spline*u_spline(nl-1))/p_spline
        end do
        ddC(nlmax,nbin)=0.
        do nl=nlmax-1,1,-1
            ddC(nl,nbin)=ddC(nl,nbin)*ddC(nl+1,nbin)+u_spline(nl)
        end do
    end do

    if (Feedback >2) write(*,*)'lmax =',ll(nlmax)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!! Compute correlation function C_p,m(theta) (using module "AngularS") !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    call Cl_to_Ctheta(ll,C,ddC,theta,dtheta,pownorm,cp,cm,nlmax,nthetamax,&
        nthetastep,ncl,xstop)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Arrange all C(theta) values in a vector, following the same order as in the data !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    nl=0
    do nr=1,2
        do nbin=1,ncl
            do k=1,nthetamax
                nl=nl+1
                if (nr==1) then
                    d_theo(nl)=cp(k,nbin)
                else
                    d_theo(nl)=cm(k,nbin)
                end if
            end do
        end do
    end do


    ! compute chi2=-2lnL corresponding to priors on fz (modulo a constant)    
    WeakLenLnLike=((CMB%norm(norm_fz)-1)/0.1)**2
    if (Feedback >2) write(*,*)'WL nuisnace parameters f_z =',CMB%norm(norm_fz)
    ! add the chi2
    do k=1,2*nthetamax*ncl
        do nk=1,2*nthetamax*ncl 
            WeakLenLnLike=WeakLenLnLike+&
                (d_theo(k)-d_obs(k))*&
                invcov(k,nk)*&
                (d_theo(nk)-d_obs(nk))
        end do
    end do

    ! compute -lnL = chi2/2
    WeakLenLnLike=WeakLenLnLike/2.

    if (Feedback >=2) write(*,*)'Weak lensing -LnLike=',WeakLenLnLike

  END FUNCTION WeakLenLnLike


end module WeakLen

