module lrg_2dCl
  use CAMB
  use CMB_Cls
  use cmbtypes
  implicit none
  
  !JD moved from lrg_pk
  integer, parameter :: lrg_maxnpoints=8192
  integer, parameter :: Nsample = 9
  real, parameter :: PI_S = 3.1415926535897932384626434E0
  integer, parameter :: nzspline = 100
  !end JD
  
  integer, parameter :: maxnz = 1000
  integer, parameter :: lspline = 70, maxl = 600 
  real :: littleh
  
!-------- Spline data for comoving distance
  real, parameter :: dz_spline = 0.01
  real :: rz(nzspline), rr(nzspline), rr2(nzspline)

  type(CAMBparams) :: P0

  type ZStruct
     integer :: nz
     real, dimension(maxnz) :: zmin, zmax
     real, dimension(1:Nsample,1:maxnz) :: zdist, zdist_abs
     real, dimension(1:Nsample) :: zmean
  end type ZStruct

  public
  private :: maxnz, littleh, P0, lspline, maxl, nzspline, dz_spline, &
       rz, rr, rr2

  contains
    
    subroutine lrg_init2d(CMB)
      type(CMBparams) :: CMB
      integer :: i
      real :: z
      

        littleh = CMB%H0/100.d0

!-------- Setup the spline for comoving distances
        do i = 1, nzspline
            rz(i) = i*dz_spline
            rr(i) = rcomov_slow(rz(i))
        end do
        call spline(rz, rr, nzspline, 1.e30, 1.e30, rr2)
         
    end subroutine lrg_init2d

    real function rcomov_slow(z)
! Return comoving distance (h^-1)
      real :: z

        rcomov_slow = DeltaTime(1._dl/(1._dl+z), 1._dl)*littleh
      return
      
    end function rcomov_slow


    real function rcomov(z)
! Use the spline if applicable, otherwise call rcomov_slow
      real :: z, r
      
        if (z .GT. rz(nzspline)) then 
            rcomov = rcomov_slow(z)
        else
            call splint(rz, rr, rr2, nzspline, z, r)
            rcomov = r
        end if

        return
    end function rcomov


    real function dcomdz(z)
! Return dy/dz
      real :: z
      real(dl) :: dtauda
      external dtauda

        dcomdz = dtauda(1._dl/(1._dl+z))*littleh/(1.0+z)**2

        return
    end function dcomdz


    subroutine lrg_rdzdist(zd, fn)
!-------- 
! Read in a redshift distribution, normalizes it and
! computes the mean
!-------- 
      type(ZStruct) :: zd
      character(*) fn
      integer :: iot, i,j,k
      real :: z1, z2, z3, z4, z5, z6, z7, z8, z9 , z10, z11, dz,z12,z13
      real, dimension (1:Nsample) :: zsum
      real :: sumzd

        zd%zmin = 0.0
        zd%zmax = 0.0
        zd%zdist = 0.0
        dz = 0.01          !the redshift bin size is hardcoded.
        open(10,file=fn, status='old')
        iot = 0
        i = 0
        do while (iot .eq. 0)
           read (10,*, iostat=iot) z1,  z3, z4, z5, z6, z7, z8, z9 , z10, z11,z12,z13
           if (iot .ne. 0) then 
                 cycle
           endif
           i = i+1
           if (i .gt. maxnz) stop 'ERROR : maxnz exceeded'
           zd%zmin(i) = z1- (dz/2.0)
           zd%zmax(i) = zd%zmin(i) + dz
           zd%zdist(1,i) = z3
           zd%zdist(2,i) = z4
           zd%zdist(3,i) = z5
           zd%zdist(4,i) = z6
           zd%zdist(5,i) = z7
           zd%zdist(6,i) = z8
           zd%zdist(7,i) = z10   !QSO0
           zd%zdist(8,i) = z11   !QSO1
           zd%zdist(9,i) = z9    ! NVSS 
        enddo
        close(10)
        if (i .eq. 0) stop 'ERROR : reading file in lrg_rdzdist'
        zd%nz = i
        do i=1, Nsample
            zsum(i) = 0.0
        enddo
        do i=1,Nsample
            do j=1,zd%nz
                zsum(i) = ((zd%zmax(j)-zd%zmin(j))*zd%zdist(i,j)) + zsum(i)
            enddo
        enddo
        zd%zdist_abs = zd%zdist

        do i=1, Nsample
            sumzd = 0.0
            zd%zdist(i,:) = zd%zdist(i,:)/zsum(i)
            do j=1,zd%nz 
                sumzd = (zd%zmax(j)-zd%zmin(j))*zd%zdist(i,j)*(zd%zmax(j)&
                    +zd%zmin(j))+ sumzd 
            enddo
            zd%zmean(i) = sumzd/2.0
            !write(*,*),"i, zmean", i, zd%zmean(i)
        enddo   

    end subroutine lrg_rdzdist

    integer function lfunc(i)
!-------- 
! The l values used for splining
!-------- 
      integer :: i, l
      real :: x

        if (i .LT. 25) then
            l = i
        else
            l = int(25.0*1.075**(i-24))
        end if

        lfunc = l
        return
    end function lfunc


    subroutine lrg_compute2dCl_Limber(CMB,Info, zd, lmax, clarr,TCMB,lsample)
!-------- 
! Code to compute 2D power spectra given a 3D power spectrum
!-------- 
      integer :: lmax
      real :: clarrtmp(lmax)
      real :: larr(lspline), cl2arr(lspline)
      real :: clarr(Nsample,lmax)
      type(ZStruct) :: zd   !normalized 
      type(CMBparams) :: CMB
      type(ParamsetInfo):: Info
      real :: clval(Nsample), w(Nsample)
      real :: rc, kh, z0, dz 
      real :: zold,bias
      real :: pkz0
      real :: rctest,chi,chiold,littleh,TCMB
      integer :: iz, il,iSa,i
      real, dimension (1:Nsample,lmax):: cl
      real:: clvaltmp
      integer, dimension(0:lmax):: lsample
      integer :: sizelsample
      
!-------- Break if lmax GT maxl (the splining breaks)
        if (lmax .GT. maxl) STOP 'ERROR! maxl exceeded'

!-------- Initialize
        call lrg_init2d(CMB)
!-------- Loop over l
        littleh = CMB%H0/100.0         !little h 


        sizelsample = lsample(0)
        do il = 1, sizelsample
            larr(il) = lsample(il)
            clval(1:Nsample) = 0.0
            w = 0.0
            zold = (zd%zmin(1)+zd%zmax(1))/4.0
            do iz = 1, zd%nz
                z0 = (zd%zmin(iz)+zd%zmax(iz))/2.0
                dz = z0-zold
                rc = com_angdist(z0,CMB) !since we can have curved universe, in Mpc/h
                kh = (larr(il)+0.5)/rc
                pkz0 = ISWgal_PowerSpecAt_Z(Info%Theory,kh,z0)
                
                w(1:Nsample) = zd%zdist_abs(1:Nsample,iz)  
                w(1:Nsample) = w(1:Nsample)*pkz0
                clval(1:Nsample) = clval(1:Nsample) +  (w(1:Nsample)*dz)
                zold = z0
            end do
            clarr(1:Nsample,il) = clval(1:Nsample)/((larr(il)+0.5)**2)*TCMB/(littleh**3)
        end do

     

    contains
      real function lrg_hubble(a,CMB)
!-------- allows matter, and Lambda, w , curved universe
!---------this is H/H0 
        implicit none
        type(CMBparams) :: CMB
        real(dl) :: a, omfac, olfac, okfac, Ea
            omfac = (CMB%omb + CMB%omc + CMB%omnu)/(a**3) !matter
            olfac = CMB%omv/(a**(3*(1+CMB%w)))   !de, w allowed
            okfac = (1.0_dl - (CMB%omb+CMB%omc+CMB%omv + CMB%omnu))/(a**2) !curvature
            Ea = omfac+okfac+olfac

            lrg_hubble = sqrt(Ea)
      end function lrg_hubble

      real(dl) function com_angdist(z,P)
!---------Calculating r: 
!---------r = 1/(\sqrt(Omega_K) H0) sin((\sqrt(Omega_K) H0) \chi) or   closed
!---------r = 1/(\sqrt(-Omega_K) H0) sinh((\sqrt(-Omega_K) H0) \chi) or  open
!---------r = \chi    flat 
        implicit none
         
        type(CMBparams) :: P
        real :: z,tmp, Mpc2km, c, hubble_0,k
            Mpc2km = 3.08568025E20
            c = 2.99792458E5/Mpc2km    ! in Mpc/s

            hubble_0 = P%H0/Mpc2km     !in sec^-1
            k = (hubble_0/c)**2*((P%omb+P%omc+P%omv+ P%omnu)-1.0_dl)
            if (k < 0.0) then 
                tmp = sqrt(-1.0*k)
                com_angdist = sinh(tmp*rcomov(z))/tmp
            else if (k > 0.0) then
                tmp = sqrt(k)
                com_angdist = sin(tmp*rcomov(z))/tmp
            else 
                com_angdist=rcomov(z)
            end if
      end function com_angdist  
       
    end subroutine lrg_compute2dCl_Limber


    

!-------- NR spline routines
    

    SUBROUTINE spline(x,y,n,yp1,ypn,y2)
      INTEGER n,NMAX
      real yp1,ypn
      real x(n), y(n), y2(n)
      PARAMETER (NMAX=10000)    ! Increased from 500 by Max
      INTEGER i,k
      real p,qn,sig,un,u(NMAX)
        if (N.gt.NMAX) pause 'SPLINE NMAX DEATH ERROR' ! Added by Max
        if (x(1).gt.x(n)) pause 'SPLINE WARNING: x NOT INCREASING' ! Added by Max
        if (yp1.gt..99e30) then
            y2(1)=0.
            u(1)=0.
        else
            y2(1)=-0.5
            u(1)=(3./(x(2)-x(1)))*((y(2)-y(1))/(x(2)-x(1))-yp1)
        endif
        do i=2,n-1
            sig=(x(i)-x(i-1))/(x(i+1)-x(i-1))
            p=sig*y2(i-1)+2.
            y2(i)=(sig-1.)/p
            u(i)=(6.*((y(i+1)-y(i))/(x(i+1)-x(i))-(y(i)-y(i-1))&
                /(x(i)-x(i-1)))/(x(i+1)-x(i-1))-sig*u(i-1))/p
        end do
        if (ypn.gt..99e30) then
            qn=0.
            un=0.
        else
            qn=0.5
            un=(3./(x(n)-x(n-1)))*(ypn-(y(n)-y(n-1))/(x(n)-x(n-1)))
        endif
        y2(n)=(un-qn*u(n-1))/(qn*y2(n-1)+1.)
        do k=n-1,1,-1
            y2(k)=y2(k)*y2(k+1)+u(k)
        end do

    END SUBROUTINE spline


    SUBROUTINE splint(xa,ya,y2a,n,x,y)
      INTEGER n
      real x,y
      real xa(n),y2a(n),ya(n)
      INTEGER k,khi,klo
      real a,b,h

!-------- If out of bounds, then linearly extrapolate
! else use the cubic spline
        if (x .LT. xa(1)) then
            y = ya(1) + (ya(1)-ya(2))*(x-xa(1))/(xa(1)-xa(2)) 
        else if (x .GT. xa(n)) then
            y = ya(n) + (ya(n)-ya(n-1))*(x-xa(n))/(xa(n)-xa(n-1))
        else
            klo=1
            khi=n
            do while (khi-klo.gt.1)
                k=(khi+klo)/2
                if(xa(k).gt.x)then
                    khi=k
                else
                    klo=k
                endif
            end do
            h=xa(khi)-xa(klo)
            if (h.eq.0.) pause 'bad xa input in splint'
            a=(xa(khi)-x)/h
            b=(x-xa(klo))/h
            y=a*ya(klo)+b*ya(khi)+((a**3-a)*y2a(klo)+(b**3-b)*y2a(khi))*(h**2)/6.
        end if

    END SUBROUTINE splint

    subroutine fortransystem(comm)
      implicit none

! This is a machine dependent implementation of SYSTEM
      character(len=*) comm

        CALL system(comm)

        return
    end subroutine fortransystem



end module lrg_2dCl
